package scanner;

import static util.Characters.*;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.function.Function;

import constructs.Token;
import constructs.TokenFactory;
import constructs.TokenFactoryImpl;
import error.LispException;
import file.FilePosition;
import file.FilePositionTracker;
import util.Characters;

/**
 * Converts a stream of bytes into a stream of Lisp tokens.
 */
public class LispScanner {

    private InputStream inputStream;
    private FilePositionTracker positionTracker;
    private TokenFactory tokenFactory;

    public LispScanner(InputStream in, String fileName) {
        this.inputStream = new LispFilterInputStream(new BufferedInputStream(in));
        this.positionTracker = new FilePositionTracker(fileName);
        this.tokenFactory = new TokenFactoryImpl();
    }

    public Token getNextToken() throws IOException {
        for (int c = inputStream.read(); c != EOF; c = inputStream.read()) {
            char currentCharacter = (char) c;
            positionTracker.incrementColumn();

            if (!Character.isWhitespace(currentCharacter))
                return createTokenFromCharacter(currentCharacter);
            else if (currentCharacter == NEWLINE)
                positionTracker.incrementLine();
        }

        return tokenFactory.createEOFToken(positionTracker.getCurrentPosition());
    }

    private Token createTokenFromCharacter(char c) throws IOException {
        FilePosition currentPosition = positionTracker.getCurrentPosition();
        String tokenText = retrieveTokenText(c);

        return tokenFactory.createToken(tokenText, currentPosition);
    }

    private String retrieveTokenText(char firstCharacter) throws IOException {
        String tokenText = "" + firstCharacter;

        if (firstCharacter == DOUBLE_QUOTE)
            tokenText = retrieveStringTokenText(firstCharacter);
        else if (Character.isDigit(firstCharacter))
            tokenText = retrieveNumberTokenText(firstCharacter);
        else if (Characters.isLegalIdentifierCharacter(firstCharacter))
            tokenText = retrieveIdentifierTokenText(firstCharacter);

        return tokenText;
    }

    private String retrieveStringTokenText(char firstDoubleQuote) throws IOException {
        ComplexTokenTextRetriever retriever = new ComplexTokenTextRetriever(firstDoubleQuote,
                                                                            Characters::isLegalStringCharacter);

        return retriever.retrieveToken();
    }

    private String retrieveNumberTokenText(char firstDigit) throws IOException {
        ComplexTokenTextRetriever retriever = new ComplexTokenTextRetriever(firstDigit, Character::isDigit);

        return retriever.retrieveToken();
    }

    private String retrieveIdentifierTokenText(char firstCharacter) throws IOException {
        ComplexTokenTextRetriever retriever = new ComplexTokenTextRetriever(firstCharacter,
                                                                            Characters::isLegalIdentifierCharacter);

        return retriever.retrieveToken();
    }

    public class ComplexTokenTextRetriever {

        Function<Character, Boolean> isPartOfToken;
        StringBuilder text;
        FilePosition position;
        char firstCharacter;
        char currentCharacter;
        char previousCharacter;

        public ComplexTokenTextRetriever(char firstCharacter, Function<Character, Boolean> isPartOfToken) {
            this.isPartOfToken = isPartOfToken;
            this.text = new StringBuilder();
            this.position = positionTracker.getCurrentPosition();
            this.firstCharacter = firstCharacter;
            this.currentCharacter = firstCharacter;
            this.previousCharacter = firstCharacter;
        }

        public String retrieveToken() throws IOException {
            text.append(firstCharacter);
            inputStream.mark(1);

            for (int c = inputStream.read(); c != EOF; c = inputStream.read()) {
                currentCharacter = (char) c;

                if (!isPartOfToken.apply(currentCharacter)) {
                    inputStream.reset();

                    return text.toString();
                }

                addCharacterToToken();

                if (isStringToken() && isTerminatingDoubleQuote())
                    return text.toString();

                previousCharacter = currentCharacter;
            }

            return terminateTokenWithEOF();
        }

        private void addCharacterToToken() {
            text.append(currentCharacter);
            positionTracker.incrementColumn();
            inputStream.mark(1);

            if (currentCharacter == NEWLINE)
                positionTracker.incrementLine();
        }

        private boolean isStringToken() {
            return firstCharacter == DOUBLE_QUOTE;
        }

        private boolean isTerminatingDoubleQuote() {
            return (currentCharacter == DOUBLE_QUOTE) && (previousCharacter != BACKSLASH);
        }

        private String terminateTokenWithEOF() {
            if (isStringToken())
                throw new UnterminatedStringException(position);

            return text.toString();
        }
    }

    public static class UnterminatedStringException extends LispException {

        private static final long serialVersionUID = 1L;
        private FilePosition position;

        public UnterminatedStringException(FilePosition position) {
            this.position = position;
        }

        @Override
        public int getSeverity() {
            return 0;
        }

        @Override
        public String getMessage() {
            return MessageFormat.format("unterminated quoted string - line {0}, column {1}", position.getLineNumber(),
                                        position.getColumnNumber());
        }
    }

}
